/***
 * mcstat - print statistics and information about a minecraft world
 * v1.0, 10-2010 by Zahl
 */

#define VERSION "1.1"

#include "helper.h"
#include "globals.h"
#include "worldloader.h"
#include "blocks.h"
#include "nbt.h"
#include <string>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#ifdef _DEBUG
#include <cassert>
#endif

using std::string;

static void chunkCallback(uint8_t* blocks, uint8_t* skylight, uint8_t* blocklight, const int offsetx, const int offsetz, bool &changed);
static void printHelp(char* binary);

static int gToY = 128, gFromY = 0;
static FILE *fileHandle = NULL, *csvFile = NULL;
static unsigned long long blockCounts[256], surfaceBlockCounts[256], totalBlockCount = 0, totalSurfaceBlockCount = 0;
static bool toSummer = false;
static bool toWinter = false;
static bool withLight = false;

int main(int argc, char** argv)
{
	// ########## command line parsing ##########
	if (argc < 2) {
		printHelp(argv[0]);
		return 1;
	}
	bool wholeworld = false;
	char *filename = NULL, *outname = NULL, *csvname = NULL;

	// First, for the sake of backward compatibility, try to parse command line arguments the old way first
#	define MOREARGS(x) (argpos + (x) < argc)
#	define NEXTARG argv[++argpos]
#	define POLLARG(x) argv[argpos + (x)]
	int argpos = 0;
	while (MOREARGS(1)) {
		const char *option = NEXTARG;
		if (strcmp(option, "-from") == 0) {
			if (!MOREARGS(2) || !isNumeric(POLLARG(1)) || !isNumeric(POLLARG(2))) {
				printf("Error: %s needs two integer arguments, ie: %s -10 5\n", option, option);
				return 1;
			}
			g_FromChunkX = atoi(NEXTARG);
			g_FromChunkZ = atoi(NEXTARG);
		} else if (strcmp(option, "-to") == 0) {
			if (!MOREARGS(2) || !isNumeric(POLLARG(1)) || !isNumeric(POLLARG(2))) {
				printf("Error: %s needs two integer arguments, ie: %s -5 20\n", option, option);
				return 1;
			}
			g_ToChunkX = atoi(NEXTARG)+1;
			g_ToChunkZ = atoi(NEXTARG)+1;
		} else if (strcmp(option, "-min") == 0) {
			if (!MOREARGS(1) || !isNumeric(POLLARG(1))) {
				printf("Error: %s needs an integer arguments, ie: %s 100\n", option, option);
				return 1;
			}
			gFromY = atoi(NEXTARG);
		} else if (strcmp(option, "-max") == 0) {
			if (!MOREARGS(1) || !isNumeric(POLLARG(1))) {
				printf("Error: %s needs an integer arguments, ie: %s 100\n", option, option);
				return 1;
			}
			gToY = atoi(NEXTARG) + 1;
		} else if (strcmp(option, "-file") == 0) {
			if (!MOREARGS(1)) {
				printf("Error: %s needs one argument, ie: %s myworld.txt\n", option, option);
				return 1;
			}
			outname = NEXTARG;
		} else if (strcmp(option, "-csv") == 0) {
			if (!MOREARGS(1)) {
				printf("Error: %s needs one argument, ie: %s world.csv\n", option, option);
				return 1;
			}
			csvname = NEXTARG;
		} else if (strcmp(option, "-summer") == 0) {
			toSummer = true;
		} else if (strcmp(option, "-winter") == 0) {
			toWinter = true;
		} else if (strcmp(option, "-light") == 0) {
			withLight = true;
		} else if (strcmp(option, "-help") == 0 || strcmp(option, "-h") == 0 || strcmp(option, "-?") == 0) {
			printHelp(argv[0]);
			return 0;
		} else {
			filename = (char*)option;
		}
	}
	wholeworld = (g_FromChunkX == UNDEFINED || g_ToChunkX == UNDEFINED);
	// ########## end of command line parsing ##########

	if (filename == NULL) {
		printf("Error: No world given. Please add the path to your world to the command line.\n");
		return 1;
	}
	if (wholeworld && !scanWorldDirectory(filename)) {
		printf("Error accessing terrain at '%s'\n", filename);
		return 1;
	}
	if (gToY > CHUNKSIZE_Y) gToY = CHUNKSIZE_Y;
	if (gFromY < 0) gFromY = 0;
	if (gToY - gFromY < 1 || g_ToChunkX <= g_FromChunkX || g_ToChunkZ <= g_FromChunkZ) {
		printf("What to doooo, yeah, what to doooo... (English: max height < 1 or X/Z-width <= 0)\n");
		return 1;
	}

	// open output file
	fileHandle = fopen((outname == NULL ? "stats.txt" : outname), "w");
	if (csvname != NULL) csvFile = fopen(csvname, "w");

	if (fileHandle == NULL && outname == NULL) {
		printf("Error opening 'stats.txt' for writing.\n");
		return 1;
	} else if (fileHandle == NULL) {
		printf("Error opening '%s' for writing.\n", outname);
		return 1;
	}

	if (csvname != NULL && csvFile == NULL) {
		printf("Error opening '%s' for writing.\n", csvname);
		return 1;
	}

	memset(blockCounts, 0, sizeof blockCounts);
	memset(surfaceBlockCounts, 0, sizeof surfaceBlockCounts);

	if (wholeworld && !loadEntireTerrain(&chunkCallback)) {
		printf("Error loading entire world.\n");
		return 1;
	} else if (!wholeworld && !loadTerrain(filename, &chunkCallback)) {
		printf("Error loading selected part of world.\n");
		return 1;
	}

	fprintf(fileHandle, "Statistics for %s:\n%d x %d x %d = %I64u blocks\n", filename,
			(g_ToChunkX - g_FromChunkX) * CHUNKSIZE_X, (g_ToChunkZ - g_FromChunkZ) * CHUNKSIZE_Z, gToY - gFromY, totalBlockCount);

	int sort[256];
	// Bubblesort should be ok for that
	for (int i = 0; i < 256; ++i) sort[i] = i;
	bool flag;
	do {
		flag = false;
		for (int i = 0; i < 255; ++i) {
			if (blockCounts[sort[i]] < blockCounts[sort[i+1]]) {
				int x = sort[i];
				sort[i] = sort[i+1];
				sort[i+1] = x;
				flag = true;
			}
		}
	} while (flag);
	fputs("\nBlock statistics:\n", fileHandle);
	for (int i = 0; i < 256; ++i) {
		if (blockCounts[sort[i]] == 0) break;
		fprintf(fileHandle, "%-" LONGEST_LENGTH "s: %6.3f%% (%I64u)\n",
				blockNames[sort[i]], (float(blockCounts[sort[i]]) / float(totalBlockCount)) * 100.0f, blockCounts[sort[i]]);
	}

	for (int i = 0; i < 256; ++i) sort[i] = i;
	do {
		flag = false;
		for (int i = 0; i < 255; ++i) {
			if (surfaceBlockCounts[sort[i]] < surfaceBlockCounts[sort[i+1]]) {
				int x = sort[i];
				sort[i] = sort[i+1];
				sort[i+1] = x;
				flag = true;
			}
		}
	} while (flag);
	fputs("\nSurface statistics:\n", fileHandle);
	for (int i = 0; i < 256; ++i) {
		if (surfaceBlockCounts[sort[i]] == 0) break;
		fprintf(fileHandle, "%-" LONGEST_LENGTH "s: %6.3f (%I64u)\n",
				blockNames[sort[i]], (float(surfaceBlockCounts[sort[i]]) / float(totalSurfaceBlockCount)) * 100.0f,
				surfaceBlockCounts[sort[i]]);
	}

	fclose(fileHandle);
	if (csvFile != NULL) fclose(csvFile);

	if (toSummer || toWinter) {
		bool ok = false;
		NBT level((string(filename) + "/level.dat").c_str(), ok);
		if (ok) {
			NBT_Tag *tData;
			ok = level.getCompound("Data", tData);
			if (ok) {
				ok = tData->setByte("SnowCovered", (toSummer ? 0 : 1));
				if (ok) printf("Season changed successfully.\n");
			}
		}
	}

	printf("Job complete.\n");
	return 0;
}

static void chunkCallback(uint8_t* blocks, uint8_t* skylight, uint8_t* blocklight, const int offsetx, const int offsetz, bool &changed)
{
	if (withLight) {
		for (int i = 0; i < 16384; ++i) {
			blocklight[i] |= skylight[i];
		}
	}
	for (int x = 0; x < CHUNKSIZE_X; ++x) {
		for (int z = 0; z < CHUNKSIZE_Z; ++z) {
			bool air = true;
			for (int y = gToY - 1; y >= gFromY; --y) {
				uint8_t &c = blocks[y + (z + x * CHUNKSIZE_Z) * CHUNKSIZE_Y];
				uint8_t l = (blocklight[(y + (z + x * CHUNKSIZE_Z) * CHUNKSIZE_Y) / 2] >> ((y % 2) * 4)) & 0x0F;
				if ((c != 0 || withLight) && csvFile != NULL) {
					if (withLight) {
						fprintf(csvFile, "%d,%d,%d,%d,%d\n", offsetx + x, y - gFromY, offsetz + z, (int)c, (int)l);
					} else {
						fprintf(csvFile, "%d,%d,%d,%d\n", offsetx + x, y - gFromY, offsetz + z, (int)c);
					}
				}
				//
				if (toSummer) {
					if (c == 78) { // SNOW
						c = 0; // AIR
						changed = true;
					} else if (c == 79) { // ICE
						c = 8; // WATER
						changed = true;
					}
				}
				//
				if (c != 0 && air) {
					air = false;
					++totalSurfaceBlockCount;
					++surfaceBlockCounts[c];
				}
				++totalBlockCount;
				++blockCounts[c];
			}
		}
	}
}

static void printHelp(char* binary)
{
	printf(
			////////////////////////////////////////////////////////////////////////////////
			"\nmcstat - convert alpha maps and get statistics - " VERSION "\n\n"
			"Usage: %s [-from X Z -to X Z] WORLDPATH\n\n"
			"  -from X Z     sets the coordinate of the chunk to start at\n"
			"  -to X Z       sets the coordinate of the chunk to end at\n"
			"                Note: Currently you need both -from and -to to define\n"
			"                bounds, otherwise the entire world will be rendered.\n"
			"  -min VAL      minimum height at which blocks will be considered (0-127)\n"
			"  -max VAL      maximum height at which blocks will be considered (0-127)\n"
			"  -file NAME    sets the output filename to 'NAME'; default is stats.txt\n"
			"  -csv NAME     write map in CSV format to 'NAME'; default is no CSV output\n"
			"  -light        add lighting information to CSV output\n"
			"  -summer       convert map to summer\n"
			"  -winter       convert map to winter\n"
			"                    Use these two at your own risk! Make a backup!\n"
			"\n    WORLDPATH is the path of the desired alpha world.\n\n"
			////////////////////////////////////////////////////////////////////////////////
			"Examples:\n\n"
#ifdef _WIN32
			"%s %%APPDATA%%\\.minecraft\\saves\\World1\n"
			"  - This would generate stats for your entire singleplayer world in slot 1\n"
			"%s -from -3 -3 -to 3 3 -csv map.csv %%APPDATA%%\\.minecraft\\saves\\World1\n"
			"  - This would generate stats for your world and also dump the map in CSV\n"
			"    format, but only from chunk (-3 -3) to chunk (3 3)\n"
#else
			"%s ~/.minecraft/saves/World1\n"
			"  - This would generate stats for your entire singleplayer world in slot 1\n"
			"%s -from -3 -3 -to 3 3 -csv map.csv ~/.minecraft/saves/World1\n"
			"  - This would generate stats for your world and also dump the map in CSV\n"
			"    format, but only from chunk (-3 -3) to chunk (3 3)\n"
#endif
			, binary, binary, binary);
}
