#include "worldloader.h"
#include "helper.h"
#include "filesystem.h"
#include "nbt.h"
#include "globals.h"
#include <list>
#include <cstring>
#include <string>
#include <cstdio>

using std::string;
typedef std::list<char*> filelist;

static filelist chunks;
static void loadChunk(const char *file, void (*callback)(uint8_t*, uint8_t*, uint8_t*, const int, const int, bool &));
static bool isAlphaWorld(string path);

bool scanWorldDirectory(const char *fromPath)
{
	if (!isAlphaWorld(string(fromPath) + "/")) {
		return false;
	}

	filelist subdirs;
	myFile file;
	DIRHANDLE d = Dir::open((char*)fromPath, file);
	if (d == NULL) {
		return false;
	}
	do {
		if (file.isdir) {
			char *s = strdup(file.name);
			subdirs.push_back(s);
		}
	} while (Dir::next(d, (char*)fromPath, file));
	Dir::close(d);
	if (subdirs.empty()) {
		return false;
	}
	// OK go
	for (filelist::iterator it = chunks.begin(); it != chunks.end(); it++) {
		free(*it);
	}
	chunks.clear();
	g_FromChunkX = g_FromChunkZ = 10000000;
	g_ToChunkX   = g_ToChunkZ  = -10000000;
	// Read subdirs now
	string base(fromPath);
	base.append("/");
	const size_t max = subdirs.size();
	size_t count = 0;
	printf("Scanning world...\n");
	for (filelist::iterator it = subdirs.begin(); it != subdirs.end(); it++) {
		string base2 = base + *it;
		printProgress(count++, max);
		d = Dir::open((char*)base2.c_str(), file);
		if (d == NULL) continue;
		do {
			if (file.isdir) {
				// Scan inside scan
				myFile chunk;
				string path = base2 + "/" + file.name;
				DIRHANDLE sd = Dir::open((char*)path.c_str(), chunk);
				if (sd != NULL) {
					do { // Here we finally arrived at the chunk files
						if (!chunk.isdir && chunk.name[0] == 'c' && chunk.name[1] == '.'
								&& chunk.name[strlen(chunk.name)-2] == 'a' && chunk.name[strlen(chunk.name)-1] == 't') { // Make sure filename is a chunk
							char *s = chunk.name;
							// Extract x coordinate from chunk filename
							s += 2;
							int valX = base10(s);
							// Extract z coordinate from chunk filename
							while (*s != '.' && *s != '\0') ++s;
							int valZ = base10(s+1);
							if (valX > -4000 && valX < 4000 && valZ > -4000 && valZ < 4000) {
								// Update bounds
								if (valX < g_FromChunkX) {
									g_FromChunkX = valX;
								}
								if (valX > g_ToChunkX) {
									g_ToChunkX = valX;
								}
								if (valZ < g_FromChunkZ) {
									g_FromChunkZ = valZ;
								}
								if (valZ > g_ToChunkZ) {
									g_ToChunkZ = valZ;
								}
								string full = path + "/" + chunk.name;
								chunks.push_back(strdup(full.c_str()));
							} else {
								printf("Ignoring bad chunk at %d %d\n", valX, valZ);
							}
						}
					} while (Dir::next(sd, (char*)path.c_str(), chunk));
					Dir::close(sd);
				}
			}
		} while (Dir::next(d, (char*)base2.c_str(), file));
		Dir::close(d);
	}
	printProgress(10, 10);
	g_ToChunkX++;
	g_ToChunkZ++;
	//
	for (filelist::iterator it = subdirs.begin(); it != subdirs.end(); it++) {
		free(*it);
	}
	printf("Min: (%d|%d) Max: (%d|%d)\n", g_FromChunkX, g_FromChunkZ, g_ToChunkX, g_ToChunkZ);
	return true;
}

bool loadEntireTerrain(void (*callback)(uint8_t*, uint8_t*, uint8_t*, const int, const int, bool &))
{
	if (chunks.empty() || callback == NULL) return false;
	const size_t max = chunks.size();
	size_t count = 0;
	printf("Loading all chunks..\n");
	for (filelist::iterator it = chunks.begin(); it != chunks.end(); it++) {
		printProgress(count++, max);
		loadChunk(*it, callback);
		free(*it);
	}
	chunks.clear();
	printProgress(10, 10);
	return true;
}

bool loadTerrain(const char* fromPath, void (*callback)(uint8_t*, uint8_t*, uint8_t*, const int, const int, bool &))
{
	if (fromPath == NULL || *fromPath == '\0' || callback == NULL) return false;
	string path(fromPath);
	if (path.at(path.size()-1) != '/') {
		path.append("/");
	}

	if (!isAlphaWorld(path)) {
		return false;
	}

	printf("Loading all chunks..\n");
	for (int chunkZ = g_FromChunkZ; chunkZ < g_ToChunkZ; ++chunkZ) {
		printProgress(chunkZ - g_FromChunkZ, g_ToChunkZ - g_FromChunkZ);
		for (int chunkX = g_FromChunkX; chunkX < g_ToChunkX; ++chunkX) {
			string thispath = path + base36((chunkX + 640000) % 64) + "/" + base36((chunkZ + 640000) % 64) + "/c." + base36(chunkX) + "." + base36(chunkZ) + ".dat";
			loadChunk(thispath.c_str(), callback);
		}
	}
	// Done loading all chunks
	printProgress(10, 10);
	return true;
}

static void loadChunk(const char *file, void (*callback)(uint8_t*, uint8_t*, uint8_t*, const int, const int, bool &))
{
	bool ok = false; // Get path name for all required chunks
	NBT chunk(file, ok);
	if (!ok) {
		return; // chunk does not exist
	}
	NBT_Tag *level = NULL;
	ok = chunk.getCompound("Level", level);
	if (!ok) {
		return;
	}
	int32_t chunkX, chunkZ;
	ok = level->getInt("xPos", chunkX);
	ok = ok && level->getInt("zPos", chunkZ);
	if (!ok) return;
	// Check if chunk is in desired bounds (not a chunk where the filename tells a different position)
	if (chunkX < g_FromChunkX || chunkX >= g_ToChunkX || chunkZ < g_FromChunkZ || chunkZ >= g_ToChunkZ) {
#ifdef _DEBUG
		printf("Chunk %s is out of bounds. %d %d\n", file, chunkX, chunkZ);
#endif
		return; // Nope, its not...
	}
	uint8_t *blockdata, *skylight, *blocklight;
	int32_t lenBlocks, lenSkylight, lenBlocklight;
	ok = level->getByteArray("Blocks", blockdata, lenBlocks)
			&& level->getByteArray("BlockLight", blocklight, lenBlocklight)
			&& level->getByteArray("SkyLight", skylight, lenSkylight);
	if (!ok || lenBlocks < 32768 || lenSkylight < 16384 || lenBlocklight < 16384) return;
	const int offsetz = (chunkZ - g_FromChunkZ) * CHUNKSIZE_Z;
	const int offsetx = (chunkX - g_FromChunkX) * CHUNKSIZE_X;

	bool doSave = false;
	(*callback)(blockdata, skylight, blocklight, offsetx, offsetz, doSave);
	if (doSave) chunk.save();
}

static bool isAlphaWorld(string path)
{
	// Check if this path is a valid minecraft world... in a pretty sloppy way
	return fileExists((path + "level.dat").c_str());
}
